/*
 * longrun(3) -- run a long-running program with checkpoints.
 *
 * We must get control to parse arguments ahead of the actual program --
 * to use longrun(3), you must modify your main program to be called _main
 * and link with longrun.o and libsave_world.a.
 *
 * The actual program must not use virtual interval timers.
 */
#ifndef lint
	static char copyright[] = "\n\
\n\
Copyright (c) 1988, 1991 by Bennet Yee.\n\
\n\
This material was developed by the above-mentioned author.  Permission to\n\
copy this software, to redistribute it, and to use it for any purpose is\n\
granted, subject to the following restrictions and understandings.\n\
\n\
1. Any copy made of this software must include this copyright notice in\n\
full.\n\
\n\
2. Users of this software agree to make their best efforts (a) to return to\n\
the above-mentioned author any improvements or extensions that they make, so\n\
that these may be included in future releases; (b) to document clearly any\n\
improvements or extensions made to this software; and (c) to inform the\n\
author of noteworthy uses of this software.  Any improvements or extensions\n\
to this software may be placed under different copyright restrictions by the\n\
author of those improvements or extensions.  That those improvements or\n\
extensions be placed under the same copyright as this software is encouraged.\n\
\n\
3. All materials containing this software or derivatives thereof must\n\
provide the contained software in source form or provide a means of\n\
obtaining the contained software in source form.\n\
\n\
4. All materials developed as a consequence of the use of this software\n\
shall duly acknowledge such use, in accordance with the usual standards of\n\
acknowledging credit in academic research.\n\
\n\
5. The author has made no warranty or representation that the operation of\n\
this software will be error-free, and the author is under no obligation to\n\
provide any services, by way of maintenance, update, or otherwise.\n\
\n\
6. In conjunction with products arising from the use of this material, there\n\
shall be no use of the names of the author, of Carnegie-Mellon University,\n\
nor of any adaptation thereof in any advertising, promotional, or sales\n\
literature without prior written consent from the author and Carnegie-Mellon\n\
University in each case.\n\
\n\
If you have questions, you can contact the author at bsy@cs.cmu.edu\n\
";
#endif /* lint */

#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>

#include "save_world.h"

#define	_LONGRUN_INTERVAL_SEC	(5*60)
#define	_LONGRUN_MIN_INTERVAL	(60)

char			*_longrun_save_name, *_longrun_pgm_name,
			*_longrun_full_pgm_name;

int			_longrun_save_fd;

char			_longrun_save_tmp[] = "longrunXXXXXXX";
char			_longrun_buf[] = "longrunDXXXXXXX";

struct itimerval	_longrun_bkup_interval = {
	_LONGRUN_INTERVAL_SEC,	0,
	_LONGRUN_INTERVAL_SEC,	0
};
int			_longrun_save_it();
struct sigvec		_longrun_timer = {
	_longrun_save_it,
	0, 0
};

void	_longrun_usage(void)
{
	fprintf(stderr,"Usage:  %s [-i interval] -s ckpoint_file -- [pgm_args]\n",_longrun_pgm_name);
	fprintf(stderr,"        or\n");
	fprintf(stderr,"        %s -r ckpoint_file (to recover)\n",_longrun_pgm_name);
}

void	_longrun_setup_save(void)
{
	if (!_longrun_save_name) {
		(void) mktemp(_longrun_buf);
		_longrun_save_name = _longrun_buf;
	}
	(void) mktemp(_longrun_save_tmp);
	_longrun_save_fd = open(_longrun_save_tmp,O_WRONLY|O_CREAT,0666);
	if (_longrun_save_fd == -1) {
		fprintf(stderr,"Can not open save-world file!\n");
		return 0;
	}
	if (sigvec(SIGVTALRM,&_longrun_timer,(struct sigvec *) 0) == -1) return 0;
	if (setitimer(ITIMER_VIRTUAL,&_longrun_bkup_interval,(struct itimerval *) 0) == -1) return 0;
	return 1;
}

void	_longrun_save_it(void)
{
	int ret;
	
	switch (ret = save_world(_longrun_save_fd,0,SAVE_ALL)) {
	case 0:
		fprintf(stderr,"save_world failed!\n");
		exit(1);
	case 1:
		if (fsync(_longrun_save_fd) == -1 || close(_longrun_save_fd) == -1) {
			perror(_longrun_pgm_name);
			fprintf(stderr,"Can not close save file\n");
			/* hope to recover info from core image -- sigh */
			abort();
		}
		if (rename(_longrun_save_tmp,_longrun_save_name) == -1) {
			perror(_longrun_pgm_name);
			fprintf(stderr,"Can not rename save file\n");
			abort();
		}
		_longrun_save_fd = open(_longrun_save_tmp,O_WRONLY|O_CREAT,0666);
		if (_longrun_save_fd == -1) {
			perror(_longrun_pgm_name);
			fprintf(stderr,"Can not open save-world file!\n");
			/* recover from core image */
			abort();
		}
		break;
	default:
#ifdef	DEBUG
		fprintf(stderr,"Restore succeeded!\n");
		fprintf(stderr,"fd was %d\n",ret);
#endif
		(void) close(ret);
		if (!_longrun_setup_save()) {
			fprintf(stderr,"Can not set up save-world!\n");
			exit(1);
		}
		break;
	}
}

void main(int	ac,
	  char	**av,
	  char	**ev)
{
	int		save_read, ch, interval, exit_val;
	extern int	_longrun_optind;
	extern char	*_longrun_optarg, *rindex();

	_longrun_full_pgm_name = av[0];
	if ((_longrun_pgm_name = rindex(av[0],'/')) == 0)
		_longrun_pgm_name = av[0];
	else	++_longrun_pgm_name;

	while ((ch = _longrun_getopt(ac,av,"i:r:s:")) != EOF) switch (ch) {
	case 'i':
		interval = atoi(_longrun_optarg);
		if (interval >= _LONGRUN_MIN_INTERVAL) {
			_longrun_bkup_interval.it_interval.tv_sec = interval;
			_longrun_bkup_interval.it_value.tv_sec = interval;
		}
		break;
	case 'r':
		/* restore! */
		save_read = open(_longrun_optarg,O_RDONLY,0);
		if (save_read == -1) {
			perror(_longrun_pgm_name);
			fprintf(stderr,"Can't open restore-world file!\n");
			exit(1);
		}
		restore_world(save_read,save_read);
		/* NOTREACHED */
		break;
	case 's':
		_longrun_save_name = _longrun_optarg;
		break;
	case '?':
		_longrun_usage();
		exit(1);
	}

	if (!_longrun_setup_save()) {
		fprintf(stderr,"Can't set up save environment!\n");
		exit(1);
	}
	ac = ac - _longrun_optind + 1;
	av += _longrun_optind - 1;
	av[0] = _longrun_full_pgm_name;

	exit_val = _main(ac,av,ev);
	/*
	 * If program calls exit itself,
	 * we won't get to unlink the temporary,
	 * so we're left with an extra checkpoint file.
	 * Big deal.
	 */

	(void) unlink(_longrun_save_tmp);
	exit(exit_val);
}
